local M = {}
-----------------------------------------------------------
-- Constants
-----------------------------------------------------------
M.FILETYPE = "toggleterm"
-- -30 and -3 is a magic number based on manual testing of what looks good
M.shading_amount = -30
M.shading_ratio = -3
-- Highlight group name prefix
M.highlight_group_name_prefix = "ToggleTerm"

return M
